#include <stdio.h>
#include <stdlib.h>
#include <time.h>
#include "cuda.h"

#define SIZE                (1024*1024)
#define THREADS_PER_BLOCK   (128)

__global__ void dotprod(float *a, float *b, float *p, int n)
{
    __shared__ float part[THREADS_PER_BLOCK];

    part[threadIdx.x] = 0.0;

    int tid = blockDim.x * blockIdx.x + threadIdx.x;
    while (tid < n)
    {
        part[threadIdx.x] += a[tid] * b[tid];
        tid += blockDim.x * gridDim.x;
    }

    __syncthreads();

	int floorPow2 = blockDim.x;
	while (floorPow2 & (floorPow2-1))
		floorPow2 &= floorPow2-1;

	if (blockDim.x != floorPow2)
	{
		if (threadIdx.x >= floorPow2)
			part[threadIdx.x - floorPow2] += part[threadIdx.x];
        __syncthreads();
	}

    int idxStep;
	for(idxStep = floorPow2 >> 1; idxStep > 32 ; idxStep >>= 1 )
	{
		if (threadIdx.x < idxStep)
			part[threadIdx.x] += part[threadIdx.x+idxStep];
        __syncthreads();
	}
	for( ; idxStep > 0 ; idxStep >>= 1 )
	{
		if (threadIdx.x < idxStep)
			part[threadIdx.x] += part[threadIdx.x+idxStep];
	}

    if (threadIdx.x == 0)
        p[blockIdx.x] = part[0];

}

int main(void)
{
    float *h_a, *h_b, *h_p;
    float *d_a, *d_b, *d_p;

    int blockspergrid = (SIZE-1)/THREADS_PER_BLOCK + 1;

	// Rezervacija pomnilnika - gostitelj
	h_a = (float *)malloc(SIZE * sizeof(float));
	h_b = (float *)malloc(SIZE * sizeof(float));
	h_p = (float *)malloc(blockspergrid * sizeof(float));

	// Inicializacija vektorjev
    srand(time(NULL));
	for (int i = 0; i < SIZE; i++)
	{
		h_a[i] = (double)rand()/RAND_MAX;
		h_b[i] = (double)rand()/RAND_MAX;;
	}

    // Rezervacija pomnilnika - naprava
    cudaMalloc((void **)&d_a, SIZE * sizeof(float));
    cudaMalloc((void **)&d_b, SIZE * sizeof(float));
    cudaMalloc((void **)&d_p, blockspergrid * sizeof(float));

    // prenos podatkov na napravo
    cudaMemcpy(d_a, h_a, SIZE * sizeof(float), cudaMemcpyHostToDevice);
    cudaMemcpy(d_b, h_b, SIZE * sizeof(float), cudaMemcpyHostToDevice);

	// Racunanje
    float elapsedTime;
    cudaEvent_t start, stop;
    cudaEventCreate(&start);
    cudaEventCreate(&stop);
    cudaEventRecord(start);

    dotprod<<<blockspergrid,THREADS_PER_BLOCK>>>(d_a, d_b, d_p, SIZE);

    cudaEventRecord(stop);
    cudaEventSynchronize(stop);
    cudaEventElapsedTime(&elapsedTime, start, stop);

    cudaError_t err = cudaGetLastError();
    if (err != cudaSuccess) 
    {
        printf("cudaGetDeviceCount error %d\n-> %s\n", err, cudaGetErrorString(err));
        exit(EXIT_FAILURE);
    }


    // prenos podatkov iz naprave
    cudaMemcpy(h_p, d_p, blockspergrid * sizeof(float), cudaMemcpyDeviceToHost);

    // sproscanje pomnilnika
    cudaFree(d_a);
    cudaFree(d_b);
    cudaFree(d_p);

    float dotProdCPU = 0.0;
    for(int i =0; i< SIZE; i++)
        dotProdCPU += h_a[i] * h_b[i];
    float dotProdGPU = 0.0;
    for(int i =0; i< blockspergrid; i++)
        dotProdGPU += h_p[i];

	// Prikaz rezultatov
    printf("dotprod(CPU): %f\ndotprod(GPU): %f\nTime(ms): %f\n", dotProdCPU, dotProdGPU, elapsedTime);

	free(h_a);
	free(h_b);
	free(h_p);

	return 0;
}
